@@ -37,6 +37,12 @@ namespace GLib {
 			g_free (ptr);
 		}
 
+		public static void Free (IntPtr[] ptrs)
+		{
+			for (int i = 0; i < ptrs.Length; i++)
+				g_free (ptrs [i]);
+		}
+
 		[DllImport("libglib-2.0-0.dll")]
 		static extern IntPtr g_filename_to_utf8 (IntPtr mem, int len, IntPtr read, out IntPtr written, out IntPtr error);
 
@@ -133,6 +139,62 @@ namespace GLib {
 				return ret.Replace ("%", "%%");
 		}
 
+		public static IntPtr[] StringArrayToNullTermPointer (string[] strs)
+		{
+			if (strs == null)
+				return new IntPtr [0];
+			IntPtr[] result = new IntPtr [strs.Length + 1];
+			for (int i = 0; i < strs.Length; i++)
+				result [i] = StringToPtrGStrdup (strs [i]);
+			result [strs.Length] = IntPtr.Zero;
+			return result;
+		}
+
+		[DllImport("libglib-2.0-0.dll")]
+		static extern void g_strfreev (IntPtr mem);
+
+		public static void StrFreeV (IntPtr null_term_array)
+		{
+			g_strfreev (null_term_array);
+		}
+
+		public static string[] NullTermPtrToStringArray (IntPtr null_term_array, bool owned)
+		{
+			if (null_term_array == IntPtr.Zero)
+				return new string [0];
+
+			int count = 0;
+			System.Collections.ArrayList result = new System.Collections.ArrayList ();
+			IntPtr s = Marshal.ReadIntPtr (null_term_array, count++ * IntPtr.Size);
+			while (s != IntPtr.Zero) {
+				result.Add (Utf8PtrToString (s));
+				s = Marshal.ReadIntPtr (null_term_array, count++ * IntPtr.Size);
+			}
+
+			if (owned)
+				g_strfreev (null_term_array);
+
+			return (string[]) result.ToArray (typeof(string));
+		}
+
+		public static string[] PtrToStringArrayGFree (IntPtr string_array)
+		{
+			if (string_array == IntPtr.Zero)
+				return new string [0];
+	
+			int count = 0;
+			while (Marshal.ReadIntPtr (string_array, count*IntPtr.Size) != IntPtr.Zero)
+				++count;
+	
+			string[] members = new string[count];
+			for (int i = 0; i < count; ++i) {
+				IntPtr s = Marshal.ReadIntPtr (string_array, i * IntPtr.Size);
+				members[i] = GLib.Marshaller.PtrToStringGFree (s);
+			}
+			GLib.Marshaller.Free (string_array);
+			return members;
+		}
+
 		// Argv marshalling -- unpleasantly complex, but
 		// don't know of a better way to do it.
 		//
@@ -145,6 +207,11 @@ namespace GLib {
 		[DllImport("libglib-2.0-0.dll")]
 		static extern IntPtr g_malloc(UIntPtr size);
 
+		public static IntPtr Malloc (ulong size)
+		{
+			return g_malloc (new UIntPtr (size));
+		}
+
 		static bool check_sixtyfour () {
 			int szint = Marshal.SizeOf (typeof (int));
 			int szlong = Marshal.SizeOf (typeof (long));
@@ -286,6 +353,10 @@ namespace GLib {
 			Array result = Array.CreateInstance (type, list.Count);
 			if (list.Count > 0)
 				list.CopyTo (result, 0);
+
+			if (type.IsSubclassOf (typeof (GLib.Opaque)))
+				list.elements_owned = false;
+
 			return result;
 		}
 	}
