@@ -31,7 +31,8 @@ namespace GLib {
 
 	public class Object : IWrapper, IDisposable {
 
-		IntPtr _obj;
+		IntPtr handle;
+		ToggleRef tref;
 		bool disposed = false;
 		Hashtable data;
 		static Hashtable Objects = new Hashtable();
@@ -40,7 +41,17 @@ namespace GLib {
 
 		~Object ()
 		{
-			Dispose ();
+			lock (PendingDestroys) {
+				lock (Objects) {
+					if (Objects[Handle] is ToggleRef)
+						PendingDestroys.Add (Objects [Handle]);
+					Objects.Remove (Handle);
+				}
+				if (!idle_queued){
+					Timeout.Add (50, new TimeoutHandler (PerformQueuedUnrefs));
+					idle_queued = true;
+				}
+			}
 		}
 
 		[DllImport("libgobject-2.0-0.dll")]
@@ -48,29 +59,18 @@ namespace GLib {
 		
 		static bool PerformQueuedUnrefs ()
 		{
-			Object [] objects;
+			object [] references;
 
 			lock (PendingDestroys){
-				objects = new Object [PendingDestroys.Count];
-				PendingDestroys.CopyTo (objects, 0);
+				references = new object [PendingDestroys.Count];
+				PendingDestroys.CopyTo (references, 0);
 				PendingDestroys.Clear ();
-			}
-			lock (typeof (Object))
 				idle_queued = false;
-
-			foreach (Object o in objects){
-				if (o._obj == IntPtr.Zero)
-					continue;
-
-				try {
-					g_object_unref (o._obj);
-				} catch (Exception e) {
-					Console.WriteLine ("Exception while disposing a " + o + " in Gtk#");
-					throw e;
-				}
-				Objects.Remove (o._obj);
-				o._obj = IntPtr.Zero;
 			}
+
+			foreach (ToggleRef r in references)
+				r.Free ();
+
 			return false;
 		}
 
@@ -80,15 +80,16 @@ namespace GLib {
 				return;
 
 			disposed = true;
-			lock (PendingDestroys){
-				PendingDestroys.Add (this);
-				lock (typeof (Object)){
-					if (!idle_queued){
-						Timeout.Add (50, new TimeoutHandler (PerformQueuedUnrefs));
-						idle_queued = true;
-					}
-				}
+			ToggleRef toggle_ref = Objects [Handle] as ToggleRef;
+			Objects.Remove (Handle);
+			try {
+				if (toggle_ref != null)
+					toggle_ref.Free ();
+			} catch (Exception e) {
+				Console.WriteLine ("Exception while disposing a " + this + " in Gtk#");
+				throw e;
 			}
+			handle = IntPtr.Zero;
 			GC.SuppressFinalize (this);
 		}
 
@@ -101,30 +102,28 @@ namespace GLib {
 				return null;
 
 			Object obj = null;
-			WeakReference weak_ref = Objects[o] as WeakReference;
 
-			if (weak_ref != null && weak_ref.IsAlive)
-				obj = weak_ref.Target as Object;
+			if (Objects.Contains (o)) {
+				ToggleRef toggle_ref = Objects [o] as ToggleRef;
+				if (toggle_ref != null && toggle_ref.IsAlive)
+					obj = toggle_ref.Target;
+			}
 
-			if (obj == null)
-				obj = Objects[o] as Object;
-
-			if (obj != null && obj._obj == o) {
-				lock (PendingDestroys)
-					PendingDestroys.Remove (obj);
+			if (obj != null && obj.Handle == o) {
 				if (owned_ref)
-					g_object_unref (obj._obj);
-				obj.disposed = false;
+					g_object_unref (obj.Handle);
 				return obj;
 			}
 
-			obj = GLib.ObjectManager.CreateObject(o); 
-			if (obj == null)
-				return null;
-
 			if (!owned_ref)
-				g_object_ref (obj.Handle);
-			Objects [o] = new WeakReference (obj);
+				g_object_ref (o);
+
+			obj = GLib.ObjectManager.CreateObject(o); 
+			if (obj == null) {
+				g_object_unref (o);
+				return null;
+			}
+
 			return obj;
 		}
 
@@ -155,16 +154,47 @@ namespace GLib {
 		private static void InvokeClassInitializers (GType gtype, System.Type t)
 		{
 			object[] parms = {gtype, t};
-			BindingFlags flags = BindingFlags.Static | BindingFlags.NonPublic | BindingFlags.FlattenHierarchy;
-			foreach (MethodInfo minfo in t.GetMethods(flags))
-				if (minfo.IsDefined (typeof (ClassInitializerAttribute), true))
-					minfo.Invoke (null, parms);
+
+			BindingFlags flags = BindingFlags.Static | BindingFlags.NonPublic;
+
+			foreach (TypeInitializerAttribute tia in t.GetCustomAttributes (typeof (TypeInitializerAttribute), true)) {
+				MethodInfo m = tia.Type.GetMethod (tia.MethodName, flags);
+				if (m != null)
+					m.Invoke (null, parms);
+			}
+
+			for (Type curr = t; curr != typeof(GLib.Object); curr = curr.BaseType) {
+ 
+				if (curr.Assembly.IsDefined (typeof (IgnoreClassInitializersAttribute), false))
+					continue;
+ 
+				foreach (MethodInfo minfo in curr.GetMethods(flags))
+					if (minfo.IsDefined (typeof (ClassInitializerAttribute), true))
+						minfo.Invoke (null, parms);
+			}
+ 		}
+
+		[DllImport("libgobject-2.0-0.dll")]
+		static extern void g_type_add_interface_static (IntPtr gtype, IntPtr iface_type, ref GInterfaceInfo info);
+
+		static void AddInterfaces (GType gtype, Type t)
+		{
+			foreach (Type iface in t.GetInterfaces ()) {
+				if (!iface.IsDefined (typeof (GInterfaceAttribute), true) || iface.IsAssignableFrom (t.BaseType))
+					continue;
+
+				GInterfaceAttribute attr = iface.GetCustomAttributes (typeof (GInterfaceAttribute), false) [0] as GInterfaceAttribute;
+				GInterfaceAdapter adapter = Activator.CreateInstance (attr.AdapterType, null) as GInterfaceAdapter;
+				
+				GInterfaceInfo info = adapter.Info;
+				g_type_add_interface_static (gtype.Val, adapter.GType.Val, ref info);
+				Console.WriteLine (adapter);
+			}
 		}
 
 		[DllImport("glibsharpglue-2")]
 		static extern IntPtr gtksharp_register_type (IntPtr name, IntPtr parent_type);
 
-
 		static int type_uid;
 		static string BuildEscapedName (System.Type t)
 		{
@@ -200,6 +230,7 @@ namespace GLib {
 			GLib.GType.Register (gtype, t);
 			ConnectDefaultHandlers (gtype, t);
 			InvokeClassInitializers (gtype, t);
+			AddInterfaces (gtype, t);
 			g_types[t] = gtype;
 			return gtype;
 		}
@@ -229,6 +260,11 @@ namespace GLib {
 			Raw = raw;
 		}
 
+		protected Object ()
+		{
+			CreateNativeObject (new string [0], new GLib.Value [0]);
+		}
+
 		[DllImport("libgobject-2.0-0.dll")]
 		static extern IntPtr g_object_new (IntPtr gtype, IntPtr dummy);
 
@@ -247,22 +283,30 @@ namespace GLib {
 			for (int i = 0; i < names.Length; i++)
 				native_names [i] = GLib.Marshaller.StringToPtrGStrdup (names [i]);
 			Raw = gtksharp_object_newv (LookupGType ().Val, names.Length, native_names, vals);
-			Objects [_obj] = this;
 			foreach (IntPtr p in native_names)
 				GLib.Marshaller.Free (p);
 		}
 
 		protected virtual IntPtr Raw {
 			get {
-				return _obj;
+				return handle;
 			}
 			set {
-				if (_obj != IntPtr.Zero)
-					Objects.Remove (_obj);
-				_obj = value;
-				if (value == IntPtr.Zero)
+				if (handle == value)
 					return;
-				Objects [value] = new WeakReference (this);
+
+				if (handle != IntPtr.Zero) {
+					Objects.Remove (handle);
+					if (tref != null) {
+						tref.Free ();
+						tref = null;
+					}
+				}
+				handle = value;
+				if (value != IntPtr.Zero) {
+					tref = new ToggleRef (this);
+					Objects [value] = tref;
+				}
 			}
 		}	
 
@@ -287,14 +331,19 @@ namespace GLib {
 			}
 		}
 
+		internal ToggleRef ToggleRef {
+			get {
+				return tref;
+			}
+		}
+
 		public IntPtr Handle {
 			get {
-				return _obj;
+				return handle;
 			}
 		}
 
 		Hashtable before_signals;
-
 		[Obsolete ("Replaced by GLib.Signal marshaling mechanism.")]
 		protected Hashtable BeforeSignals {
 			get {
@@ -339,15 +388,19 @@ namespace GLib {
 
 		void NotifyCallback (IntPtr handle, IntPtr pspec, IntPtr gch)
 		{
-			GLib.Signal sig = ((GCHandle) gch).Target as GLib.Signal;
-			if (sig == null)
-				throw new Exception("Unknown signal GC handle received " + gch);
+			try {
+				GLib.Signal sig = ((GCHandle) gch).Target as GLib.Signal;
+				if (sig == null)
+					throw new Exception("Unknown signal GC handle received " + gch);
 
-			NotifyArgs args = new NotifyArgs ();
-			args.Args = new object[1];
-			args.Args[0] = pspec;
-			NotifyHandler handler = (NotifyHandler) sig.Handler;
-			handler (GLib.Object.GetObject (handle), args);
+				NotifyArgs args = new NotifyArgs ();
+				args.Args = new object[1];
+				args.Args[0] = pspec;
+				NotifyHandler handler = (NotifyHandler) sig.Handler;
+				handler (GLib.Object.GetObject (handle), args);
+			} catch (Exception e) {
+				ExceptionManager.RaiseUnhandledException (e, false);
+			}
 		}
 
 		void ConnectNotification (string signal, NotifyHandler handler)
@@ -387,7 +440,6 @@ namespace GLib {
 			return Handle.GetHashCode ();
 		}
 
-		[Obsolete("Can cause instability due to garbage collection of GLib.Objects.")]
 		public Hashtable Data {
 			get { 
 				if (data == null)
@@ -397,9 +449,13 @@ namespace GLib {
 			}
 		}
 
+		Hashtable persistent_data;
 		protected Hashtable PersistentData {
 			get {
-				return WeakObject.Lookup (Handle).Data;
+				if (persistent_data == null)
+					persistent_data = new Hashtable ();
+				
+				return persistent_data;
 			}
 		}
 
